// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "ms.h"

#include <assert.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>

static size_t ms_sod(struct ms_gr const *gr) {
  assert(gr->ne >= 0);
  assert(gr->ne <= INT_MAX / 2 - 1);
  int nd = (1 + gr->ne) * 2;
  assert((unsigned)nd <= (SIZE_MAX - sizeof *gr) / sizeof *gr->d);
  return (unsigned)nd * sizeof *gr->d;
}

struct ms_gr *ms_rdgr(FILE *fp) {
  struct ms_gr grh;
  size_t ct = fread(&grh, sizeof grh, 1, fp);
  assert(ct);
  size_t sod = ms_sod(&grh);
  struct ms_gr *gr = malloc(sizeof *gr + sod);
  assert(gr);
  *gr = grh;
  ct = fread(&gr->d, sod, 1, fp);
  assert(ct);
  return gr;
}

void ms_wrgr(struct ms_gr const *gr, FILE *fp) {
  size_t ct = fwrite(gr, sizeof *gr + ms_sod(gr), 1, fp);
  assert(ct);
}

void ms_xgr(struct ms_gr *gr) {
  assert(gr->ne >= 0);
  assert(gr->ne <= INT_MAX / 2 - 1);
  int nd = (1 + gr->ne) * 2;
  assert((unsigned)nd <= SIZE_MAX);
  signed char *bd;
  MS_CALLOC(bd, (unsigned)nd);
  int nc = 0;
  int *st;
  MS_CALLOC(st, (unsigned)(nd - 2));
  int nf = 0;
  for (int di = 2; di < nd; di++) {
    if (bd[di]) {
      continue;
    }
    nc++;
    int nst = 0;
    int dj = di;
    goto x;
    do {
      dj = st[--nst];
      if (bd[dj]) {
        continue;
      }
     x:
      ;
      int dk = dj;
      do {
        assert(!bd[dk]);
        bd[dk] = 1;
        st[nst++] = dk ^ 1;
        gr->d[dk].fi = nf;
        dk = gr->d[dk].vdi ^ 1;
        assert(dk >= 2);
        assert(dk < nd);
      } while (dk != dj);
      nf++;
    } while (nst);
  }
  free(st);
  free(bd);
  gr->nc = nc;
  gr->nf = nf;
}

void ms_ckgr(struct ms_gr const *gr) {
  assert(gr->nc >= 0);
  assert(gr->nc <= INT_MAX / 2);
  assert(gr->nv >= 0);
  assert((unsigned)gr->nv <= SIZE_MAX);
  assert(gr->ne >= 0);
  assert(gr->ne <= INT_MAX / 2 - 1);
  int nd = (1 + gr->ne) * 2;
  assert((unsigned)nd <= SIZE_MAX);
  assert(gr->nf >= 0);
  assert((unsigned)gr->nf <= SIZE_MAX);
  assert(gr->nv - gr->ne == gr->nc * 2 - gr->nf);
  // vdi
  signed char *bd;
  MS_CALLOC(bd, (unsigned)nd);
  for (int di = 2; di < nd; di++) {
    int vdi = gr->d[di].vdi;
    assert(vdi >= 2);
    assert(vdi < nd);
    assert(!bd[vdi]);
    bd[vdi] = 1;
  }
  // nc
  for (int di = 2; di < nd; di++) {
    bd[di] = 0;
  }
  int nc = 0;
  int *st;
  MS_CALLOC(st, (unsigned)(nd - 2));
  for (int di = 2; di < nd; di++) {
    if (bd[di]) {
      continue;
    }
    nc++;
    int nst = 0;
    int dj = di;
    goto x;
    do {
      dj = st[--nst];
      if (bd[dj]) {
        continue;
      }
     x:
      ;
      int dk = dj;
      do {
        bd[dk] = 1;
        st[nst++] = dk ^ 1;
        dk = gr->d[dk].vdi;
      } while (dk != dj);
    } while (nst);
  }
  free(st);
  assert(gr->nc == nc);
  // vi
  for (int di = 2; di < nd; di++) {
    bd[di] = 0;
  }
  signed char *bv;
  MS_CALLOC(bv, (unsigned)gr->nv);
  for (int di = 2; di < nd; di++) {
    if (bd[di]) {
      continue;
    }
    bd[di] = 1;
    int vi = gr->d[di].vi;
    assert(vi >= 0);
    assert(vi < gr->nv);
    assert(!bv[vi]);
    bv[vi] = 1;
    for (int dj = gr->d[di].vdi; dj != di; dj = gr->d[dj].vdi) {
      assert(gr->d[dj].vi == vi);
      bd[dj] = 1;
    }
  }
  for (int vi = 0; vi < gr->nv; vi++) {
    assert(bv[vi]);
  }
  free(bv);
  // fi
  for (int di = 2; di < nd; di++) {
    bd[di] = 0;
  }
  signed char *bf;
  MS_CALLOC(bf, (unsigned)gr->nf);
  for (int di = 2; di < nd; di++) {
    if (bd[di]) {
      continue;
    }
    bd[di] = 1;
    int fi = gr->d[di].fi;
    assert(fi >= 0);
    assert(fi < gr->nf);
    assert(!bf[fi]);
    bf[fi] = 1;
    for (int dj = gr->d[di].vdi ^ 1; dj != di; dj = gr->d[dj].vdi ^ 1) {
      assert(gr->d[dj].fi == fi);
      bd[dj] = 1;
    }
  }
  for (int fi = 0; fi < gr->nf; fi++) {
    assert(bf[fi]);
  }
  free(bf);
  free(bd);
  // l
  for (int di = 2; di < nd; di++) {
    int l = gr->d[di].l;
    assert(l >= 0);
    assert(gr->d[di ^ 1].l <= INT_MAX - l);
  }
}
