// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "common.h"

#include <stdlib.h>

static size_t compute_size(size_t struct_size,
                           uint64_t num_elements,
                           size_t element_size) {
  assert(num_elements <= (SIZE_MAX - struct_size) / element_size);
  return (size_t)(struct_size + num_elements * element_size);
}

void *allocate(size_t size) {
  if (!size) {
    return NULL;
  }
  void *ptr = malloc(size);
  if (!ptr) {
    perror(NULL);
    assert(0);
  }
  return ptr;
}

void *allocate_array(size_t struct_size,
                     uint64_t num_elements,
                     size_t element_size) {
  return allocate(compute_size(struct_size, num_elements, element_size));
}

void deallocate(void *ptr) {
  free(ptr);
}

void *reallocate(void *ptr, size_t size) {
  if (!size) {
    free(ptr);
    return NULL;
  }
  ptr = realloc(ptr, size);
  if (!ptr) {
    perror(NULL);
    assert(0);
  }
  return ptr;
}

void *reallocate_array(void *ptr,
                       size_t struct_size,
                       uint64_t num_elements,
                       size_t element_size) {
  return reallocate(ptr, compute_size(struct_size, num_elements, element_size));
}
