// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include <assert.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>

typedef char Assume_eight_bit_char[(CHAR_BIT == 8) * 2 - 1];

struct Point {
  int32_t x;
  int32_t y;
};

struct Line {
  struct Point p;
  struct Point q;
};

struct Line_list {
  int64_t num_lines;
  int64_t max_lines;
  struct Line line[];
};

// allocate.c
void *allocate(size_t size);
void *allocate_array(size_t struct_size,
                     uint64_t num_elements,
                     size_t element_size);
void deallocate(void *ptr);
void *reallocate(void *ptr, size_t size);
void *reallocate_array(void *ptr,
                       size_t struct_size,
                       uint64_t num_elements,
                       size_t element_size);

// decode_coordinate.c
int32_t decode_coordinate(uint64_t coord);

// euclidean_length.c
int64_t euclidean_length(struct Point p, struct Point q);

// line_list.c
void line_list_initialize(struct Line_list **line_list_ptr);
void line_list_append(struct Line line, struct Line_list **line_list_ptr);

// orientation_less.c
int orientation_less(struct Line l, struct Line m);

// read_shapefiles.c
void read_shapefiles(FILE *stream, struct Line_list **line_list_ptr);

// z_order.c
uint64_t expand(int32_t x);
int32_t unexpand(uint64_t n);
uint64_t z_rank(struct Point p);
struct Point z_unrank(uint64_t n);
