// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "common.h"

#define EXP2(n) (UINT64_C(1) << (n))
#define MASK(n) (EXP2(n) - 1)
#define EXTRACT(dest, n, src) \
  dest = src & MASK(n);       \
  src >>= (n)

int32_t decode_coordinate(uint64_t coord) {
  static int const s_num_bits = 1;
  static int const q_num_bits = 11;
  static int const c_hi_num_bits = 20;
  static int const c_lo_num_bits = 32;
  static uint64_t const precision = 1000000;
  EXTRACT(uint32_t c_lo, c_lo_num_bits, coord);
  EXTRACT(uint32_t c_hi, c_hi_num_bits, coord);
  uint64_t carry = precision * c_lo;
  EXTRACT(uint32_t p_frac2, c_lo_num_bits, carry);
  uint64_t p_whole = carry + precision * (c_hi | EXP2(c_hi_num_bits));
  EXTRACT(int q, q_num_bits, coord);
  int r = c_hi_num_bits - (q - 1023);
  assert(r > 0);
  if (r > 64) {
    return 0;
  }
  EXTRACT(uint64_t p_frac1, r - 1, p_whole);
  EXTRACT(int p_frac0, 1, p_whole);
  p_whole += p_frac0 && (p_frac1 || p_frac2 || (p_whole & 1));
  EXTRACT(int s, s_num_bits, coord);
  int64_t x = (int64_t)p_whole;
  if (s) {
    x = -x;
    assert(x >= INT32_MIN);
  } else {
    assert(x <= INT32_MAX);
  }
  return (int32_t)x;
}
