// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "ms.h"

#include <assert.h>
#include <stdlib.h>

struct ms_sd {
  int vdi;
  int l;
  signed char b;
  signed char q;
};

int *ms_mkmst(struct ms_gr const *gr) {
  int *pr;
  MS_CALLOC(pr, (unsigned)gr->nv);
  int nd = (1 + gr->ne) * 2;
  struct ms_sd *sd;
  MS_CALLOC(sd, (unsigned)nd);
  for (int di = 2; di < nd; di++) {
    sd[di].vdi = gr->d[di].vdi;
    sd[di].l = gr->d[di].l + gr->d[di ^ 1].l;
  }
  int *st;
  MS_CALLOC(st, (unsigned)(nd - 2));
  int nst = 0;
  for (int di = 2; di < nd; di++) {
    st[nst++] = di;
  }
  while (nst) {
    for (int sti = 0; sti < nst; sti++) {
      int di = st[sti];
      if (sd[di].b) {
        continue;
      }
      sd[di].b = 1;
      int am = di;
      int m = sd[am].l;
      signed char q = 2;
      for (int dj = sd[di].vdi; dj != di; dj = sd[dj].vdi) {
        sd[dj].b = 1;
        if (sd[dj].l < m || (sd[dj].l == m && (dj >> 1) < (am >> 1))) {
          am = dj;
          m = sd[am].l;
          q = 2;
        } else if (sd[dj].l == m && (dj >> 1) == (am >> 1)) {
          q = 1;
        }
      }
      sd[am].q = q;
      sd[am ^ 1].q = q;
    }
    for (int sti = 0; sti < nst; sti++) {
      int di = st[sti];
      if (!sd[di].b) {
        continue;
      }
      sd[di].b = 0;
      int am = di;
      int m = sd[am].l;
      signed char q = 1;
      for (int dj = sd[di].vdi ^ 1; dj != di; dj = sd[dj].vdi ^ 1) {
        sd[dj].b = 0;
        if (sd[dj].l > m || (sd[dj].l == m && (dj >> 1) > (am >> 1))) {
          am = dj;
          m = sd[am].l;
          q = 1;
        } else if (sd[dj].l == m && (dj >> 1) == (am >> 1)) {
          q = 2;
        }
      }
      sd[am].q = q;
      sd[am ^ 1].q = q;
    }
    int stj = 0;
    for (int sti = 0; sti < nst; sti++) {
      int di = st[sti];
      if (sd[di].q) {
        continue;
      }
      st[stj++] = di;
      int dj = sd[di].vdi;
      if (!sd[dj].q) {
        continue;
      }
      do {
        dj = sd[dj ^ (sd[dj].q >> 1)].vdi;
      } while (sd[dj].q);
      sd[di].vdi = dj;
    }
    nst = stj;
  }
  for (int di = 2; di < nd; di++) {
    int vi = gr->d[di].vi;
    if (pr[vi]) {
      continue;
    }
    pr[vi] = 1;
    nst = 0;
    int dj = di;
    goto x;
    do {
      dj = st[--nst];
      vi = gr->d[dj].vi;
      if (pr[vi]) {
        continue;
      }
      pr[vi] = dj;
     x:
      ;
      int dk = dj;
      do {
        if (sd[dk].q >> 1) {
          st[nst++] = dk ^ 1;
        }
        dk = gr->d[dk].vdi;
      } while (dk != dj);
    } while (nst);
  }
  for (int vi = 0; vi < gr->nv; vi++) {
    if (pr[vi] == 1) {
      pr[vi] = 0;
    }
  }
  free(st);
  free(sd);
  return pr;
}

struct ms_m {
  int ei;
  int ll;
};

static struct ms_m ms_dm(struct ms_gr const *gr, int di) {
  return (struct ms_m){di >> 1, gr->d[di].l + gr->d[di ^ 1].l};
}

static int ms_mltm(struct ms_m m, struct ms_m n) {
  return m.ll < n.ll || (m.ll == n.ll && m.ei < n.ei);
}

void ms_ckmst(struct ms_gr const *gr, int const *pr) {
  ms_ckpr(gr, pr);
  int *p2;
  MS_CALLOC(p2, (unsigned)gr->nv);
  for (int vi = 0; vi < gr->nv; vi++) {
    p2[vi] = pr[vi];
  }
  int nd = (1 + gr->ne) * 2;
  for (int di = 2; di < nd; di += 2) {
    int vi = gr->d[di].vi;
    int vj = gr->d[di ^ 1].vi;
    if (pr[vi] == di || pr[vj] == (di ^ 1)) {
      continue;
    }
    struct ms_m m = ms_dm(gr, di);
    int dj = 1;
    while (1) {
      int dk = p2[vi];
      p2[vi] = dj ^ 1;
      if (!dk) {
        break;
      }
      dj = dk;
      vi = gr->d[dj ^ 1].vi;
    }
    while (1) {
      int dk = p2[vj];
      if (!dk) {
        break;
      }
      assert(ms_mltm(ms_dm(gr, dk), m));
      vj = gr->d[dk ^ 1].vi;
    }
  }
  free(p2);
}
