// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "ms.h"

#include <assert.h>
#include <limits.h>
#include <stdlib.h>

struct ms_h {
  int pr;
  unsigned k;
};

static void ms_ins(struct ms_h *h, int ii, struct ms_h i) {
  h->k = i.k;
  while (1) {
    int ij = ii >> 1;
    if (i.k - h[ij].k <= INT_MAX) {
      break;
    }
    h[ii] = h[ij];
    ii = ij;
  }
  h[ii] = i;
}

int *ms_mksp(struct ms_gr const *gr, unsigned **kp) {
  int *pr;
  MS_CALLOC(pr, (unsigned)gr->nv);
  unsigned *k;
  MS_CALLOC(k, (unsigned)gr->nv);
  int nd = (1 + gr->ne) * 2;
  struct ms_h *h;
  MS_CALLOC(h, (unsigned)nd);
  for (int di = 2; di < nd; di++) {
    int vi = gr->d[di].vi;
    if (pr[vi]) {
      continue;
    }
    pr[vi] = 1;
    int nh = 0;
    struct ms_h i = {di, 0};
    int dj = i.pr;
    goto x;
    do {
      h[nh + 1].k = h[nh].k;
      int ii = 1;
      i = h[ii];
      while (ii <= (nh >> 1)) {
        int ij = ii * 2;
        if (h[ij + 1].k - h[ij].k > INT_MAX) {
          ij++;
        }
        h[ii] = h[ij];
        ii = ij;
      }
      if (ii < nh) {
        ms_ins(h, ii, h[nh]);
      }
      nh--;
      vi = gr->d[i.pr].vi;
      if (pr[vi]) {
        continue;
      }
      pr[vi] = i.pr;
      dj = gr->d[i.pr].vdi;
     x:
      k[vi] = i.k;
      do {
        int dk = dj ^ 1;
        ms_ins(h, ++nh, (struct ms_h){dk, i.k + (unsigned)gr->d[dk].l});
        dj = gr->d[dj].vdi;
      } while (dj != i.pr);
    } while (nh);
  }
  free(h);
  for (int vi = 0; vi < gr->nv; vi++) {
    if (pr[vi] == 1) {
      pr[vi] = 0;
    }
  }
  *kp = k;
  return pr;
}

void ms_cksp(struct ms_gr const *gr, int const *pr, unsigned const *k) {
  ms_ckpr(gr, pr);
  for (int vi = 0; vi < gr->nv; vi++) {
    int di = pr[vi];
    if (di) {
      assert(k[gr->d[di ^ 1].vi] + (unsigned)gr->d[di].l == k[vi]);
    } else {
      assert(!k[vi]);
    }
  }
  int nd = (1 + gr->ne) * 2;
  for (int di = 2; di < nd; di++) {
    assert(k[gr->d[di ^ 1].vi] + (unsigned)gr->d[di].l - k[gr->d[di].vi]
           <= INT_MAX);
  }
}

unsigned *ms_mkk(struct ms_gr const *gr, int const *pr) {
  unsigned *k;
  MS_CALLOC(k, (unsigned)gr->nv);
  int *t = ms_mkt(gr, pr);
  for (int ti = 0; ti < gr->nv; ti++) {
    int vi = t[ti];
    int di = pr[vi];
    if (di) {
      k[vi] = k[gr->d[di ^ 1].vi] + (unsigned)gr->d[di].l;
    }
  }
  free(t);
  return k;
}
