// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

struct ms_f {
  struct ms_f *p;
  struct ms_f *l;
  struct ms_f *r;
  int dc;
  int dm;
  int di;
  int ll;
};

static void ms_1up(struct ms_f const *u, int *dm) {
  if (u && u->dm + u->dc < *dm) {
    *dm = u->dm + u->dc;
  }
}

#define MS_UP(L, R)                        \
  static void ms_##L##up(struct ms_f *u) { \
    int dm = 0;                            \
    ms_1up(u->L, &dm);                     \
    ms_1up(u->R, &dm);                     \
    u->dm = dm;                            \
  }

MS_UP(l, r)
MS_UP(r, l)

#define MS_ROT(L, R)                                        \
  static void ms_##L##rot(struct ms_f *u, struct ms_f *v) { \
    v->R = u->L;                                            \
    u->L = v;                                               \
    v->p = u;                                               \
    v->dc -= u->dc;                                         \
    if (v->R) {                                             \
      v->R->p = v;                                          \
      v->R->dc -= v->dc;                                    \
    }                                                       \
    ms_##R##up(v);                                          \
  }

MS_ROT(l, r)
MS_ROT(r, l)

#define MS_2ROT(L, R)       \
  w = v->p;                 \
  if (w) {                  \
    struct ms_f *x = w->p;  \
    if (w->R == v) {        \
      v->dc += w->dc;       \
      u->dc += v->dc;       \
      ms_##L##rot(v, w);    \
      ms_##L##rot(u, v);    \
      v = x;                \
      continue;             \
    } else if (w->L == v) { \
      v->dc += w->dc;       \
      u->dc += v->dc;       \
      ms_##L##rot(u, v);    \
      ms_##R##rot(u, w);    \
      v = x;                \
      continue;             \
    }                       \
  }                         \
  u->p = w;                 \
  u->dc += v->dc;           \
  ms_##L##rot(u, v);        \
  break

static void ms_splf(struct ms_f *u) {
  struct ms_f *v = u->p;
  struct ms_f *w = u;
  while (1) {
    if (v) {
      if (v->r == w) {
        MS_2ROT(l, r);
      } else if (v->l == w) {
        MS_2ROT(r, l);
      }
    }
    u->p = v;
    break;
  }
}

static void ms_expf(struct ms_f *u) {
  ms_splf(u);
  struct ms_f *v = u->p;
  u->p = 0;
  while (v) {
    ms_splf(v);
    if (v->l) {
      v->l->dc += v->dc;
    }
    struct ms_f *w = v->p;
    ms_rrot(u, v);
    v = w;
  }
}

static void ms_ebtf(struct ms_f *u, struct ms_f *v) {
  int dc = u->dc;
  if (u->r) {
    u->r->dc += dc;
  }
  if (u->l) {
    u->l->dc += dc;
    ms_ebtf(u->l, v);
  }
  dc = v->ll - v->dc;
  v->dc = u->dc;
  u->dc = dc;
  int di = v->di;
  v->di = u->di;
  u->di = di ^ 1;
  int ll = v->ll;
  v->ll = u->ll;
  u->ll = ll;
  if (u->r) {
    ms_ebtf(u->r, v);
    u->r->dc -= dc;
  }
  if (u->l) {
    u->l->dc -= dc;
  }
  struct ms_f *l = u->l;
  u->l = u->r;
  u->r = l;
  ms_lup(u);
}

static void ms_inif(struct ms_f *u) {
  u->p = 0;
  u->l = 0;
  u->r = 0;
}

static void ms_lnkf(struct ms_f *u,
                    struct ms_f *v,
                    int m,
                    int di,
                    int ll) {
  ms_expf(u);
  if (u->l) {
    u->l->dc += u->dc - m;
  }
  u->dc = m;
  ms_lup(u);
  u->di = di;
  u->ll = ll;
  ms_expf(v);
  int dc = v->dc;
  if (v->l) {
    v->l->dc += dc;
  }
  v->l = u;
  u->p = v;
  u->dc -= dc;
}

static int ms_pvtf(struct ms_f *u,
                   struct ms_f *v,
                   int di,
                   int ll) {
  if (u == v) {
    return di;
  }
  ms_expf(v);
  if (v->r) {
    int dc = v->dc;
    if (v->l) {
      v->l->dc += dc;
    }
    v->r->dc += dc;
    ms_ebtf(v->r, v);
    v->r->dc -= ll;
    v->l = v->r;
    v->r = 0;
  } else if (v->l) {
    v->l->dc += v->dc - ll;
  }
  v->dc = ll;
  v->di = di;
  v->ll = ll;
  ms_expf(u);
  if (u->l) {
    u->l->dc += u->dc;
    u->l = 0;
  }
  struct ms_f *w;
  if (u->r->dm + u->r->dc >= 0) {
    w = u;
  } else {
    w = u->r;
    while (1) {
      if (w->l && w->l->dm + w->l->dc == w->dm) {
        w = w->l;
      } else if (!w->dm) {
        break;
      } else {
        w = w->r;
      }
    }
    ms_splf(w);
  }
  w->dc = 0;
  if (w != v) {
    if (w->r) {
      w->r->p = 0;
      w->r = 0;
    }
    ms_splf(v);
    ms_lup(v);
    ms_splf(u);
    int dc = u->dc;
    if (u->l) {
      u->l->dc += dc;
    }
    u->l = v;
    v->p = u;
    v->dc -= dc;
  }
  return w->di;
}
