// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "submain.h"

#include <assert.h>
#include <stdlib.h>
#include "../ms.h"
#include "dt.c"

enum {
  DL = 0
};

static void bc_update_c(struct ms_gr const *gr, int const *pr, int vi,
                        unsigned long long *c) {
  for (int vj = 0; vj < gr->nv; vj++) {
    int vk = vj;
    while (1) {
      int di = pr[vk];
      if (!di) {
        break;
      }
      vk = gr->d[di ^ 1].vi;
    }
    if (vk != vi) {
      continue;
    }
    vk = vj;
    while (1) {
      c[vk]++;
      int di = pr[vk];
      if (!di) {
        break;
      }
      vk = gr->d[di ^ 1].vi;
    }
    c[vk]--;
    c[vi]--;
  }
  c[vi]++;
}

unsigned long long submain(struct ms_gr const *gr) {
  unsigned *k;
  int *pr = ms_mksp(gr, &k);
  struct ms_f *f = ms_mkf(gr, pr, k);
  free(k);
  int *mstpr = ms_mkmst(gr);
  int *t = ms_mkt(gr, mstpr);
  free(mstpr);
  struct bc_v *v;
  MS_CALLOC(v, (unsigned)gr->nv);
  for (int vi = 0; vi < gr->nv; vi++) {
    bc_initialize(&v[vi]);
  }
  struct bc_args args = {gr, v, 0};
  for (int vi = 0; vi < gr->nv; vi++) {
    bc_on_pivot(&args, 0, pr[vi]);
  }
  args.np = 0;
  unsigned long long *c = 0;
  if (DL >= 2) {
    MS_CALLOC(c, (unsigned)gr->nv);
  }
  for (int ti = 0; ti < gr->nv; ti++) {
    int vi = t[ti];
    ms_chr(gr, pr, f, bc_on_pivot, &args, vi);
    bc_new_root(&v[vi]);
    if (c) {
      bc_update_c(gr, pr, vi, c);
    }
  }
  if (DL >= 3) {
    for (int vi = 0; vi < gr->nv; vi++) {
      printf("%d: %llu\n", vi, bc_centrality(&v[vi]));
    }
  }
  if (c) {
    for (int vi = 0; vi < gr->nv; vi++) {
      assert(bc_centrality(&v[vi]) == c[vi]);
    }
    free(c);
  }
  free(v);
  free(t);
  free(f);
  free(pr);
  return args.np;
}
