// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "submain.h"

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include "../ms.h"
#include "../set_affinity/set_affinity.h"
#include "dt.c"

enum {
  DL = 0
};

static int parseint(char const *s) {
  int i = 0;
  do {
    if (i) {
      assert(*s >= '0');
    } else {
      assert(*s >= '1');
    }
    assert(*s <= '9');
    int d = *s - '0';
    assert(i < INT_MAX / 10 || (i == INT_MAX / 10 && d <= INT_MAX % 10));
    i = i * 10 + d;
  } while (*++s);
  return i;
}

static struct ms_gr const *gr;
static int denom;
static unsigned long long *c;
static unsigned long long np;
static int *t;
static int ti;
static pthread_mutex_t mutex;

static void *start_routine(void *arg) {
  unsigned *k;
  int *pr = ms_mksp(gr, &k);
  struct ms_f *f = ms_mkf(gr, pr, k);
  free(k);
  struct bc_v *v;
  MS_CALLOC(v, (unsigned)gr->nv);
  for (int vi = 0; vi < gr->nv; vi++) {
    bc_initialize(&v[vi]);
  }
  struct bc_args args = {gr, v, 0};
  for (int vi = 0; vi < gr->nv; vi++) {
    bc_on_pivot(&args, 0, pr[vi]);
  }
  args.np = 0;
  int err = pthread_mutex_lock(&mutex);
  assert(!err);
  while (1) {
    int tj = ti;
    if (tj >= gr->nv) {
      break;
    }
    int tl = tj + (gr->nv - tj - 1) / denom + 1;
    ti = tl;
    err = pthread_mutex_unlock(&mutex);
    assert(!err);
    for (int tk = tj; tk < tl; tk++) {
      int vi = t[tk];
      ms_chr(gr, pr, f, bc_on_pivot, &args, vi);
      bc_new_root(&v[vi]);
    }
    err = pthread_mutex_lock(&mutex);
    assert(!err);
  }
  for (int vi = 0; vi < gr->nv; vi++) {
    c[vi] += bc_centrality(&v[vi]);
  }
  np += args.np;
  err = pthread_mutex_unlock(&mutex);
  assert(!err);
  free(v);
  free(f);
  free(pr);
  return 0;
}

unsigned long long submain(struct ms_gr const *gr1) {
  gr = gr1;
  char const *s = getenv("NUM_THREADS");
  assert(s);
  int num_threads = parseint(s);
  assert(num_threads <= INT_MAX / 2);
  denom = 2 * num_threads;
  MS_CALLOC(c, (unsigned)gr->nv);
  np = 0;
  int *mstpr = ms_mkmst(gr);
  t = ms_mkt(gr, mstpr);
  free(mstpr);
  ti = 0;
  int err = pthread_mutex_init(&mutex, (pthread_mutexattr_t const *)0);
  assert(!err);
  pthread_t *thread;
  MS_CALLOC(thread, (unsigned)num_threads);
  pthread_attr_t attr;
  err = pthread_attr_init(&attr);
  assert(!err);
  errno = 0;
  long num_processors = sysconf(_SC_NPROCESSORS_ONLN);
  assert(!errno);
  for (int i = 0; i < num_threads; i++) {
    if (num_threads <= num_processors) {
      long q = num_processors / num_threads;
      long r = num_processors % num_threads;
      set_affinity(&attr, i < r ? i * q + i + q : i * q + q + r - 1);
    }
    err = pthread_create(&thread[i], &attr, start_routine, (void *)0);
    assert(!err);
  }
  err = pthread_attr_destroy(&attr);
  assert(!err);
  for (int i = 0; i < num_threads; i++) {
    void *value;
    err = pthread_join(thread[i], &value);
    assert(!err);
  }
  free(thread);
  err = pthread_mutex_destroy(&mutex);
  assert(!err);
  free(t);
  if (DL >= 3) {
    for (int vi = 0; vi < gr->nv; vi++) {
      printf("%d: %llu\n", vi, c[vi]);
    }
  }
  free(c);
  return np;
}
